/*
  Servo.h - Interrupt driven Servo library for Arduino using 16 bit timers- Version 2
  Copyright (c) 2009 Michael Margolis.  All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*
  A servo is activated by creating an instance of the Servo class passing
  the desired pin to the attach() method.
  The servos are pulsed in the background using the value most recently
  written using the write() method.

  The methods are:

    Servo - Class for manipulating servo motors connected to Arduino pins.

    attach(pin )  - Attaches a servo motor to an i/o pin.
    attach(pin, min, max  ) - Attaches to a pin setting min and max values in microseconds
    default min is 544, max is 2400

    write()     - Sets the servo angle in degrees.  (invalid angle that is valid as pulse in microseconds is treated as microseconds)
    writeMicroseconds() - Sets the servo pulse width in microseconds
    read()      - Gets the last written servo pulse width as an angle between 0 and 180.
    readMicroseconds()   - Gets the last written servo pulse width in microseconds.
    attached()  - Returns true if there is a servo attached.
    detach()    - Stops an attached servos from pulsing its i/o pin.
 */

#ifndef Servo_h
#define Servo_h

#include <inttypes.h>

#define usToTicks(_us)    (clockCyclesPerMicrosecond() * _us) // converts microseconds to ticks
#define ticksToUs(_ticks) (( (unsigned)_ticks ) / clockCyclesPerMicrosecond()) // converts from ticks back to microseconds

#define SERVO_MIN() (MIN_PULSE_WIDTH - this->min * 4)  // minimum value in uS for this servo
#define SERVO_MAX() (MAX_PULSE_WIDTH - this->max * 4)  // maximum value in uS for this servo

#define MIN_PULSE_WIDTH       544    // the shortest pulse sent to a servo
#define MAX_PULSE_WIDTH      2400    // the longest pulse sent to a servo
#define DEFAULT_PULSE_WIDTH  1500    // default pulse width when servo is attached
#define REFRESH_INTERVAL    20000    // minumim time to refresh servos in microseconds

#define MAX_SERVOS             12    // maximum servos allowed

#define INVALID_SERVO         255    // flag indicating an invalid servo index

#define TRIM_DURATION           3    // compensation ticks for digitalWrite

typedef struct  {
  uint32_t nbr        :6 ;           // a pin number from 0 to 63
  uint32_t isActive   :1 ;           // true if this channel is enabled, pin not pulsed if false
} ServoPin_t;

typedef struct {
  ServoPin_t Pin;
  volatile unsigned int ticks;
} servo_t;

class Servo
{
public:
  Servo();
  uint32_t attach(int pin);          // attach the given pin to the next free channel, sets pinMode, returns channel number or 0 if failure
  uint32_t attach(int pin, int min, int max); // as above but also sets min and max values for writes.
  void detach();
  void write(int value);             // if value is < 200 its treated as an angle, otherwise as pulse width in microseconds
  void writeMicroseconds(int value); // Write pulse width in microseconds
  int read();                        // returns current pulse width as an angle between 0 and 180 degrees
  int readMicroseconds();            // returns current pulse width in microseconds for this servo
  bool attached();                   // return true if this servo is attached, otherwise false
private:
  uint32_t servoIndex;               // index into the channel data for this servo
  int32_t min;                       // minimum is this value times 4 added to MIN_PULSE_WIDTH
  int32_t max;                       // maximum is this value times 4 added to MAX_PULSE_WIDTH
};

#endif
